import 'package:contentsquare/src/modules/screengraph/screengraph_utils.dart';
import 'package:flutter/material.dart';
import 'package:contentsquare_platform_interface/contentsquare_platform_interface.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';
import 'package:webview_flutter/webview_flutter.dart';

/// The format of the background color used here is: #OORRGGBB
/// OO: Opacity
/// RR: Red
/// GG: Green
/// BB: Blue
const _defaultElementBackgroundColor = '#00FFFFFF';

/// Represent a single node in the screengraph model
class ScreengraphNode {
  /// Flutter [Element] associated with the current [ScreengraphNode]
  final Element element;

  /// Index of element in the list `parent.children`
  final int index;

  /// Parent [ScreengraphNode]
  ScreengraphNode? _parent;

  /// Parent of this node
  ScreengraphNode? get parent => _parent;

  /// children of this node
  final List<ScreengraphNode> children;

  /// ScreengraphNode
  ScreengraphNode({required Element element, required int index, ScreengraphNode? parent, required List<ScreengraphNode> children})
      : this._(element, index, parent, children);

  ScreengraphNode._(this.element, this.index, this._parent, this.children) {
    for (final it in children) {
      it._parent = this;
    }
  }

  String _getPath({String rootPath = ""}) {
    final parent = this.parent;
    final className = element._widgetName;
    if (parent != null) {
      return '${parent._getPath(rootPath: rootPath)}>$className:eq($index)';
    } {
      return '$rootPath>$className:eq($index)';
    }
  }

  /// Path to the current node
  /// it has empty `rootPath`
  String get path {
    return _getPath();
  }

  /// Format current [ScreengraphNode] to a JSON map
  Future<Map> toJson({required String rootPath, required double pixelRatio}) async {
    final frame = element._frame;
    final bitmap = await element._bitmap;
    final List childrenJson = [];
    for (final child in children) {
      childrenJson.add(await child.toJson(rootPath: rootPath, pixelRatio: pixelRatio));
    }
    return {
      'id': '${element.widget.key}',
      'metadata': {
        'class_name': element.widget.runtimeType.toString(),
        'fullpath': _getPath(rootPath: rootPath),
      },

      /// "recordingId" should only mandatory for session replay
      /// On the iOS SDK, the structure is used for both: snapshoting and session replay features
      /// so it ends up mandatory for the screengraph passed down to the ExternalBridgeInterface.takeSnapshot completion
      /// For this reason, we set temporarily a 'recordingId' value based on the widget's hascode.
      /// TODO: This will need to be checked and discussed when implementing session replay related features
      'recordingId': element.widget.hashCode,
      'style': {
        'height': frame.height * pixelRatio,
        'width': frame.width * pixelRatio,
        'x': frame.left * pixelRatio,
        'y': frame.top * pixelRatio,
        'visibility': true,
        if (bitmap != null) 'bmp': bitmap,

        if (bitmap == null) 'bg': element._backgroundColor,
      },
      'format': element._viewFormat.toInt(),
      'children': childrenJson,
    };
  }
}

/// Extension method to get children.
extension GetChildren on Element {

  /// getChildren
  List<_IndexedElement> getChildren() {
    var childElements = toDiagnosticsNode().getChildren().map((it) => it.value as Element).toList();
    if (this is SliverMultiBoxAdaptorElement) {
      final List<_IndexedElement> indexedElements = [];
      for (final child in childElements) {
        final renderObject = child.renderObject;
        final indexedObject = renderObject is RenderIndexedSemantics ? renderObject : null;
        indexedElements.add(_IndexedElement(child, indexedObject?.index ?? 0));
      }
      return indexedElements;
    }
    List<Element> adjustedChildren = [];
    for (final child in childElements) {
      if (_shouldIgnoreChild(child)) {
        final nested = child.getChildren().map((it) => it.element);
        adjustedChildren.addAll(nested);
      }  else {
        adjustedChildren.add(child);
      }
    }
    final List<_IndexedElement> indexedElements = [];
    for (var i = 0; i < adjustedChildren.length; i++) {
      indexedElements.add(_IndexedElement(adjustedChildren[i], i));
    }
    return indexedElements;
  }

  bool _shouldIgnoreChild(Element child) {
    // If this child has same frame as parent and doesn't have any useful information, ignore this child.
    final isChildBlank = child._frame == _frame && !hasElementBitmap(child) && !child._hasBackgroundColor;
    // We don't want to ignore recyclable widget (ListView, GridView, SliverList, SliverGrid...), otherwise its child's index will be wrong.
    return child is! SliverMultiBoxAdaptorElement && isChildBlank || child._isPrivate;
  }
}

extension on Element {
  /// Returns the frame Rect (position + size) of the element
  Rect get _frame {
    final renderObject = findRenderObject();
    if (renderObject is RenderView) {
      final renderBox = renderObject.child;
      if (renderBox == null) {
        return Rect.zero;
      }
      return getFrameForBox(renderBox);
    } else if (renderObject is RenderBox) {
      return getFrameForBox(renderObject);
    } else if (renderObject is RenderSliver) {
      return getFrameForSliver(renderObject);
    } else if (renderObject is RenderAbstractViewport) {
      return getFrameForViewport(renderObject);
    } else {
      return Rect.zero;
    }
  }

  /// TODO: Retrieve the actual background color of the current element
  /// This requires to analyze the widget type and provide a way to retrieve the color for each type of widget
  String get _backgroundColor {
    return _defaultElementBackgroundColor;
  }

  bool get _hasBackgroundColor {
    return _backgroundColor != _defaultElementBackgroundColor;
  }

  CSViewFormat get _viewFormat {
    /// For now we only support [CSViewFormat.mobileView] & [CSViewFormat.webViewContainer]
    /// TODO: Add support for CSViewFormat.webElement (https://contentsquare.atlassian.net/browse/FLUT-23)
    return (widget is WebView
        ? CSViewFormat.webViewContainer
        : CSViewFormat.mobileView);
  }

  Future<String?> get _bitmap async {
    return getElementBitmap(this);
  }

  String get _widgetName {
    return trimGenericParams(widget.runtimeType.toString());
  }

  bool get _isPrivate {
    return widget.runtimeType.toString().startsWith('_');
  }
}

class _IndexedElement {
  final Element element;
  final int index;

  _IndexedElement(this.element, this.index);
}