import 'dart:convert';

import 'package:flutter/rendering.dart';
import 'package:flutter/widgets.dart';
import 'dart:ui' as ui;

/// trimGenericParams
///
/// Returns the given class name without generic params
String trimGenericParams(String className) {
  return className.replaceAll(RegExp(r'\<.*\>'), '');
}

Rect _getFrameForRenderObject(RenderObject renderObject) {
  final translation = renderObject.getTransformTo(null).getTranslation();
  return renderObject.paintBounds.shift(Offset(translation.x, translation.y));
}

/// getFrameForBox
///
/// Returns the location Rect (position + size) of the given [RenderBox]
Rect getFrameForBox(RenderBox renderBox) {
  if (!renderBox.hasSize) {
    return Rect.zero;
  }
  return _getFrameForRenderObject(renderBox);
}

/// getFrameForSliver
///
/// Returns the location Rect (position + size) of the given [RenderSliver]
Rect getFrameForSliver(RenderSliver renderSliver) {
  if (renderSliver.geometry == null) {
    return Rect.zero;
  }
  return _getFrameForRenderObject(renderSliver);
}

/// getFrameForViewport
///
/// Returns the location Rect (position + size) of the given [RenderAbstractViewport]
Rect getFrameForViewport(RenderAbstractViewport renderViewport) {
  return _getFrameForRenderObject(renderViewport);
}

/// getElementBitmap
///
/// Returns the bitmap image of the given element
Future<String?> getElementBitmap(Element element) async {
  if (element.findRenderObject() is RenderRepaintBoundary) {
    final renderRepaintBoundary =
        (element.findRenderObject() as RenderRepaintBoundary);

    ui.Image? image;

    try {
      if (!_canCastToImage(renderRepaintBoundary)) {
        return null;
      }
      image = await renderRepaintBoundary.toImage();
      final bytes = (await image.toByteData(format: ui.ImageByteFormat.png))
          ?.buffer
          .asUint8List();
      if (bytes != null) {
        return base64.encode(bytes);
      }
    } catch (_) {
    } finally {
      image?.dispose();
    }
  }
  return null;
}

/// Returns if the element has bitmap image
bool hasElementBitmap(Element element) {
  if (element.findRenderObject() is RenderRepaintBoundary) {
    final renderRepaintBoundary =
    (element.findRenderObject() as RenderRepaintBoundary);
    return _canCastToImage(renderRepaintBoundary);
  }
  return false;
}

/// renderRepaintBoundary widgets can find themselves in states where they cannot be rasterized to images
/// when the paint state is dirty or when the size is 0 along one of the axis
/// for instance when they descend from a Scrollable
bool _canCastToImage(RenderRepaintBoundary renderRepaintBoundary) {
  return (renderRepaintBoundary.size.width > 0 &&
      renderRepaintBoundary.size.height > 0);
}
