import 'package:contentsquare/src/modules/gestures/cs_gesture_detector.dart';
import 'package:contentsquare_platform_interface/contentsquare_platform_interface.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:mocktail/mocktail.dart';
import '../../../helpers/matchers.dart';

abstract class StaticGestureCallableClass {
  call(Offset position);
}

abstract class DynamicGestureCallableClass {
  call(
    Offset position,
    CSGestureData data,
  );
}

class MockStaticGestureCallback extends Mock
    implements StaticGestureCallableClass {}

class MockDynamicGestureCallback extends Mock
    implements DynamicGestureCallableClass {}

class FakeGCGestureData extends Fake implements CSGestureData {}

void main() {
  late MockStaticGestureCallback mockStaticGestureCallback;
  late MockDynamicGestureCallback mockDynamicGestureCallback;

  setUpAll(() {
    registerFallbackValue(FakeGCGestureData());
  });

  setUp(() {
    mockStaticGestureCallback = MockStaticGestureCallback();
    mockDynamicGestureCallback = MockDynamicGestureCallback();
  });

  group('CSGestureDetector', () {
    testWidgets('should call "onTap" callback when "tap" gesture is performed',
        (WidgetTester tester) async {
      await tester.pumpWidget(
        CSGestureDetector(
          onTap: mockStaticGestureCallback,
          child: const Center(
            child: SizedBox(
              width: 200,
              height: 200,
            ),
          ),
        ),
      );

      /// [WidgetTester] root size is (w: 800, h: 600)
      /// A [Center]ed [SizedBox] is therefore position at:
      /// Rect(top: 300, left: 200, width: 200, height: 200)
      const tapPosition = Offset(378.0, 313.0);
      await tester.tapAt(tapPosition);
      verify(() => mockStaticGestureCallback(tapPosition));
    });

    testWidgets(
        'should call "onTap" callback when moving gesture is performed within a distance < [kTouchSlop]',
        (WidgetTester tester) async {
      await tester.pumpWidget(
        CSGestureDetector(
          onTap: mockStaticGestureCallback,
          child: const Center(
            child: SizedBox(
              width: 200,
              height: 200,
            ),
          ),
        ),
      );

      const point = Offset(378.0, 313.0);
      const delta = Offset(0, 17); // 17 < kTouchSlop == 18
      await tester.dragFrom(point, delta);
      verify(() => mockStaticGestureCallback(point));
    });

    testWidgets(
        'should call "onLongPress" callback when "long press" gesture is performed',
        (WidgetTester tester) async {
      await tester.pumpWidget(CSGestureDetector(
        onLongPress: mockStaticGestureCallback,
        child: Container(),
      ));
      const point = Offset(154.0, 386.0);
      await tester.longPressAt(point);

      verify(() => mockStaticGestureCallback(point));
    });

    testWidgets(
        'should call "onDrag" callback with the right direction, distance and 0 velocity when "drag" gesture is performed upward',
        (WidgetTester tester) async {
      await tester.pumpWidget(CSGestureDetector(
        onDrag: mockDynamicGestureCallback,
        child: Container(),
      ));
      const dragFrom = Offset(200, 200);
      await tester.dragFrom(
        dragFrom,
        const Offset(0, -50),
      );

      const expectedGestureData = CSGestureData(
        direction: CSGestureDirection.up,
        distance: 50,
        // velocity is not computed for "drag" gesture (only for flick)
        velocity: 0,
      );
      expect(
        verify(
          () => mockDynamicGestureCallback(
            dragFrom,
            captureAny(),
          ),
        ).captured.single,
        almostEqualGesture(expectedGestureData),
      );
    });

    testWidgets(
        'should call "onDrag" callback with the right direction, distance and 0 velocity when "drag" gesture is performed rightward',
        (WidgetTester tester) async {
      await tester.pumpWidget(CSGestureDetector(
        onDrag: mockDynamicGestureCallback,
        child: Container(),
      ));
      const dragFrom = Offset(200, 200);
      await tester.dragFrom(
        dragFrom,
        const Offset(75, 0),
      );

      const expectedGestureData = CSGestureData(
        direction: CSGestureDirection.right,
        distance: 75,
        // velocity is not computed for "drag" gesture (only for flick)
        velocity: 0,
      );
      expect(
        verify(
          () => mockDynamicGestureCallback(
            dragFrom,
            captureAny(),
          ),
        ).captured.single,
        almostEqualGesture(expectedGestureData),
      );
    });

    testWidgets(
        'should call "onFlick" callback with the right direction, distance and velocity when "flick" gesture is performed downward',
        (WidgetTester tester) async {
      await tester.pumpWidget(CSGestureDetector(
        onFlick: mockDynamicGestureCallback,
        child: Container(),
      ));
      const flickFrom = Offset(200, 200);
      await tester.timedDragFrom(
        flickFrom,
        const Offset(0, 100),
        const Duration(milliseconds: 250),
      );

      const expectedGestureData = CSGestureData(
        distance: 100,
        direction: CSGestureDirection.down,
        velocity: 400,
      );
      expect(
        verify(
          () => mockDynamicGestureCallback(
            flickFrom,
            captureAny(),
          ),
        ).captured.single,
        almostEqualGesture(expectedGestureData),
      );
    });

    testWidgets(
        'should call "onFlick" callback with the right direction, distance and velocity when "flick" gesture is performed leftward',
        (WidgetTester tester) async {
      await tester.pumpWidget(CSGestureDetector(
        onFlick: mockDynamicGestureCallback,
        child: Container(),
      ));
      const flickFrom = Offset(200, 200);
      await tester.timedDragFrom(
        flickFrom,
        const Offset(-100, 0),
        const Duration(milliseconds: 500),
      );

      const expectedGestureData = CSGestureData(
        distance: 100,
        direction: CSGestureDirection.left,
        velocity: 200,
      );
      expect(
        verify(
          () => mockDynamicGestureCallback(
            flickFrom,
            captureAny(),
          ),
        ).captured.single,
        almostEqualGesture(expectedGestureData),
      );
    });
  });
}
