import 'dart:convert';

import 'package:contentsquare/src/modules/screengraph/screengraph.dart';
import 'package:flutter/material.dart';
import 'package:flutter_test/flutter_test.dart';

import 'fixtures/gesture_path.dart';
import 'fixtures/screengraph1.dart';

void main() {

  Future<void> setUpWidget(WidgetTester tester) async {
    tester.binding.window.physicalSizeTestValue = const Size(300, 300);
    await tester.pumpWidget(Directionality(
      textDirection: TextDirection.ltr,
      child: MaterialApp(
        home: Padding(
            padding: EdgeInsets.zero,
            child: Padding(
              padding: const EdgeInsets.all(1),
              child: Scaffold(
                body: ListView.builder(itemBuilder: (context, index) {
                  return ListTile(
                    title: SizedBox(
                        height: 50,
                        width: 50,
                        child: Text('Text $index', key: ValueKey('Item$index'),)
                    ),
                  );
                }, itemCount: 20,),
              ),
            )
        ),
      ),
    ));
    final itemFinder = find.byKey(const ValueKey('Item15'));
    await tester.scrollUntilVisible(itemFinder, 100, scrollable: find.byType(Scrollable));
  }
  
  group('Screengraph Generator', () {
    group('generateScreengraph', () {
      testWidgets(
          'should return a screengraph with fullpath containing no private widgets',
          (WidgetTester tester) async {
        await tester.pumpWidget(const Directionality(
          textDirection: TextDirection.ltr,
          child: Material(),
        ));

        //With private widget skipping: FakeRootPath>RenderObjectToWidgetAdapter:eq(0)>Directionality:eq(0)>Material:eq(0)>AnimatedPhysicalModel:eq(0)>PhysicalModel:eq(0)>NotificationListener:eq(0)
        //Without private widget skipping: FakeRootPath>RenderObjectToWidgetAdapter:eq(0)>Directionality:eq(0)>Material:eq(0)>AnimatedPhysicalModel:eq(0)>PhysicalModel:eq(0)>NotificationListener:eq(0)>_InkFeatures:eq(0)

        var screengraph = await Screengraph()
            .toJson("FakeRootPath", pixelRatio: 1.0);
        expect(
            jsonEncode(screengraph),
            isNot(contains("_InkFeatures")));
      });

      testWidgets(
          'should return correct screen graph json',
              (WidgetTester tester) async {
            await setUpWidget(tester);

            final expectedJson = jsonDecode(screengraph1);

            await tester.runAsync(() async {
              var screengraph = await Screengraph()
                  .toJson("FakeRootPath", pixelRatio: 1.0);
              _updateJson(screengraph, "bmp", "bmp");
              _updateJson(screengraph, "id", "id");
              _updateJson(screengraph, "recordingId", "recordingId");
              expect(screengraph, expectedJson);
            });
          });

      testWidgets(
          'should return correct gesture path',
              (WidgetTester tester) async {
            await setUpWidget(tester);
            const position = Offset(2, 2);
            final gesturePath = Screengraph.atPosition(position).pathAtPosition(position);
            expect(screengraph1, contains(gesturePath));
            expect(gesturePath, gesturePath1);
          });

      testWidgets(
          'should return correct gesture path for unhitable widget',
              (WidgetTester tester) async {
            await setUpWidget(tester);
            final renderObject = find.byKey(const ValueKey('Item15')).evaluate().firstWhere((it) => it.renderObject is RenderBox).renderObject as RenderBox;
            final position = renderObject.localToGlobal(const Offset(40, 20));
            final gesturePath = Screengraph.atPosition(position).pathAtPosition(position);
            expect(screengraph1, contains(gesturePath));
            expect(gesturePath, gesturePath2);
          });
    });
  });
}

void _updateJson(dynamic json, String key, dynamic value) {
  if (json is List) {
    for (final it in json) {
      _updateJson(it, key, value);
    }
  } else if (json is Map) {
    if (json[key] != null) {
      json[key] = value;
    }
    _updateJson(json.values.toList(), key, value);
  }
}